from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Agg')

from toy_problem_mnist import fmnist,mnist,mnist_perf
fmnist_L_x,fmnist_L_y = fmnist(True)
fmnist_T_x,fmnist_T_y = fmnist(False)

def lerp(a,b,λ): return λ*b + (1-λ)*a

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

def analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=128,batch_size=64,perf_batch_size=None,fashion=True,mini_data=False):
	if perf_batch_size==None: perf_batch_size=batch_size
	model.to(device)
	model.identifier=identifier
	optim = torch.optim.Adam(model.parameters(),lr=10**lr0)
	hist = {
		"epoch_L":[], "loss_L": [], "loss_L_μ": [], "loss_L_σ": [], "perf_L_precision": [], "perf_L_accuracy": [], "perf_L_recall": [], "perf_L_F1": [],
		"epoch_T":[], "loss_T": [], "loss_T_μ": [], "loss_T_σ": [], "perf_T_precision": [], "perf_T_accuracy": [], "perf_T_recall": [], "perf_T_F1": [],
	}
	# if fashion:
	# 	dataset_L = torch.utils.data.DataLoader(list(zip(fmnist_L_x,fmnist_L_y)), batch_size=batch_size, shuffle=True)
	# 	dataset_T = torch.utils.data.DataLoader(list(zip(fmnist_T_x,fmnist_T_y)), batch_size=batch_size, shuffle=True)
	# else:
	# 	dataset_L = torch.utils.data.DataLoader(list(zip( mnist_L_x, mnist_L_y)), batch_size=batch_size, shuffle=True)
	# 	dataset_T = torch.utils.data.DataLoader(list(zip( mnist_T_x, mnist_T_y)), batch_size=batch_size, shuffle=True)
	if fashion:
		dataset_L_x = fmnist_L_x
		dataset_L_y = fmnist_L_y
		dataset_T_x = fmnist_T_x
		dataset_T_y = fmnist_T_y
	else:
		dataset_L_x = mnist_L_x
		dataset_L_y = mnist_L_y
		dataset_T_x = mnist_T_x
		dataset_T_y = mnist_T_y
	if mini_data==True: mini_data=1000
	if not mini_data==False:
		dataset_L_x = dataset_L_x[:mini_data]
		dataset_L_y = dataset_L_y[:mini_data]
		dataset_T_x = dataset_T_x[:mini_data]
		dataset_T_y = dataset_T_y[:mini_data]
	dataset_L = torch.utils.data.DataLoader(list(zip( dataset_L_x, dataset_L_y)), batch_size=batch_size, shuffle=True, pin_memory=True)
	dataset_T = torch.utils.data.DataLoader(list(zip( dataset_T_x, dataset_T_y)), batch_size=batch_size, shuffle=True, pin_memory=True)
	for epoch in range(epochs):
		lr = 10**lerp(lr0,lr1,epoch/(epochs-1))
		for g in optim.param_groups: g['lr'] = lr
		#
		model.train()
		losses=[]
		for i,(x,y) in enumerate(dataset_L):
			# print("hey I'm a debug message and you should delete me !")
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			optim.zero_grad();loss.requires_grad_().backward();optim.step()
			losses.append(loss.item())
			hist["loss_L"].append(loss.item())
			hist["epoch_L"].append(epoch + i*batch_size/fmnist_L_x.shape[0])
		hist["loss_L_μ"].append(np.mean(losses))
		hist["loss_L_σ"].append(np.std(losses))
		#
		model.eval()
		losses=[]
		for i,(x,y) in enumerate(dataset_T):
			# print("hey I'm another debug message and you should delete me !")
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			losses.append(loss.item())
			hist["loss_T"].append(loss.item())
			hist["epoch_T"].append(epoch + i*batch_size/fmnist_T_x.shape[0])
		hist["loss_T_μ"].append(np.mean(losses))
		hist["loss_T_σ"].append(np.std(losses))
		#
		rgb_L = (0,.5,1)
		rgb_T = (1,.5,0)
		#
		# cm_L,ac_L,pr_L,re_L,F1_L=mnist_perf(model,device,"learn",fmnist_L_x,fmnist_L_y,2048,rgb_L) ### on learn set
		# cm_L,ac_L,pr_L,re_L,F1_L=mnist_perf(model,device,"learn",fmnist_L_x,fmnist_L_y,batch_size,rgb_L) ### on learn set
		cm_L,ac_L,pr_L,re_L,F1_L=mnist_perf(model,device,"learn",fmnist_L_x[:10000],fmnist_L_y[:10000],perf_batch_size,rgb_L) ### on learn set
		hist["perf_L_precision"].append(pr_L)
		hist["perf_L_recall"   ].append(re_L)
		hist["perf_L_accuracy" ].append(ac_L)
		hist["perf_L_F1"       ].append(F1_L)
		#
		# cm_T,ac_T,pr_T,re_T,F1_T=mnist_perf(model,device,"test" ,fmnist_T_x,fmnist_T_y,2048,rgb_T) ### on test set
		cm_T,ac_T,pr_T,re_T,F1_T=mnist_perf(model,device,"test" ,fmnist_T_x,fmnist_T_y,perf_batch_size,rgb_T) ### on test set
		hist["perf_T_precision"].append(pr_T)
		hist["perf_T_recall"   ].append(re_T)
		hist["perf_T_accuracy" ].append(ac_T)
		hist["perf_T_F1"       ].append(F1_T)
		os.makedirs(f"out/",exist_ok=True)
		# print(f"epoch={epoch+1:3d}/{epochs:3d} | log(loss)={np.log(hist['loss_L_μ'][-1]):.5f} | L_prec={pr_L:.5f} L_recall={re_L:.5f} L_accur={ac_L:.5f} L_F1={F1_L:.5f} | T_prec={pr_T:.5f} T_recall={re_T:.5f} T_accur={ac_T:.5f} T_F1={F1_T:.5f}")
		print(f"epoch={epoch+1:3d}/{epochs:3d} | log(loss)={np.log(hist['loss_L_μ'][-1]):.5f} | L_accur={ac_L:.5f} | T_accur={ac_T:.5f}")
		#
		plt.figure()
		plt.scatter(hist["epoch_L"],np.log(hist['loss_L']), label='log(loss) train', s=0.05, alpha=1/4, color=rgb_L)
		plt.scatter(hist["epoch_T"],np.log(hist['loss_T']), label='log(loss) test' , s=0.05, alpha=1/4, color=rgb_T)
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss.png");plt.close('all')
		#
		plt.figure()
		plt.plot(hist['perf_L_accuracy' ], label=f'learn accuracy' , color=rgb_L , linestyle='-')
		plt.plot(hist['perf_L_precision'], label=f'learn precision', color=rgb_L, linestyle='--')
		plt.plot(hist['perf_L_recall'   ], label=f'learn recall'   , color=rgb_L, linestyle=':') # TODO : why perf_L_recall == perf_L_accuracy ?
		plt.plot(hist['perf_L_F1'       ], label=f'learn F1'       , color=rgb_L, linestyle='-.')
		plt.plot(hist['perf_T_accuracy' ], label=f'test accuracy'  , color=rgb_T, linestyle='-')
		plt.plot(hist['perf_T_precision'], label=f'test precision' , color=rgb_T, linestyle='--')
		plt.plot(hist['perf_T_recall'   ], label=f'test recall'    , color=rgb_T, linestyle=':') # TODO : why perf_T_recall == perf_T_accuracy ?
		plt.plot(hist['perf_T_F1'       ], label=f'test F1'        , color=rgb_T, linestyle='-.')
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-perf.png");plt.close('all')
		#
		plt.figure()
		plt.plot(range(epoch+1),hist['loss_L_μ'], label='learn loss', color=rgb_L)
		plt.plot(range(epoch+1),hist['loss_T_μ'], label='test loss' , color=rgb_T)
		plt.fill_between(range(epoch+1), np.array(hist['loss_L_μ'])-np.array(hist['loss_L_σ']), np.array(hist['loss_L_μ'])+np.array(hist['loss_L_σ']), alpha=0.2, color=rgb_L)
		plt.fill_between(range(epoch+1), np.array(hist['loss_T_μ'])-np.array(hist['loss_T_σ']), np.array(hist['loss_T_μ'])+np.array(hist['loss_T_σ']), alpha=0.2, color=rgb_T)
		plt.legend()
		plt.yscale('log')
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss-2.png");plt.close('all')
		#
		with h5py.File(f'out/{model.identifier}.h5', 'w') as file:
			for k in hist: file.create_dataset(k, data=hist[k])

# def analysis_auto_enc(model,identifier,lr0,lr1,epochs=64,batch_size=64):
# 	pass # TODO

# for loss_name,loss in [
# 		["MSE",nn.MSELoss()],
# 	]:
# 	for σ_name,σ in σs:
# 		for N in [4]:
# 			for m_size in [2]:
# 				for M in [1,2,4,8,16]:
# 					for ker_size in [3]:
# 						for dropout in [0,0.25,0.5,0.75]:
# 							add_job(lambda: analysis_classification(
# 								nn_cn(N,m_size,M,ker_size),
# 								f"cn({N},{m_size},{M},{ker_size})-{σ_name}-dropout={dropout}-loss={loss_name}",
# 								loss
# 							))
